"""DataCite writer for commonmeta-py"""
import orjson as json
from typing import Optional, Union

from ..base_utils import wrap, compact
from ..doi_utils import doi_from_url
from ..constants import (
    CM_TO_BIB_TRANSLATIONS,
    CM_TO_CSL_TRANSLATIONS,
    CM_TO_CR_TRANSLATIONS,
    CM_TO_DC_TRANSLATIONS,
    CM_TO_RIS_TRANSLATIONS,
    CM_TO_SO_TRANSLATIONS,
    Commonmeta,
)


def write_datacite(metadata: Commonmeta) -> Optional[Union[str, dict]]:
    """Write datacite. Make sure JSON Schema validates before writing"""
    if metadata.write_errors is not None:
        return "{}"

    creators = [
        to_datacite_creator(i)
        for i in wrap(metadata.contributors)
        if i.get("contributorRoles", None) == ["Author"]
    ]
    contributors = [
        to_datacite_creator(i)
        for i in wrap(metadata.contributors)
        if i.get("contributorRoles", None) != ["Author"]
    ]
    related_identifiers = [
        to_datacite_related_identifier(i)
        for i in wrap(metadata.references)
        if i.get("doi", None) or i.get("url", None)
    ]

    resource__typegeneral = CM_TO_DC_TRANSLATIONS.get(metadata.type, "Other")
    resource_type = CM_TO_CR_TRANSLATIONS.get(metadata.type, "Other")
    if resource__typegeneral == resource_type or resource__typegeneral in [
        "Dataset",
        "JournalArticle",
        "Other",
        "Preprint",
        "Software",
    ]:
        resource_type = None
    types = compact(
        {
            "resourceTypeGeneral": resource__typegeneral,
            "resourceType": resource_type,
            "schemaOrg": CM_TO_SO_TRANSLATIONS.get(metadata.type, "CreativeWork"),
            "citeproc": CM_TO_CSL_TRANSLATIONS.get(metadata.type, "article"),
            "bibtex": CM_TO_BIB_TRANSLATIONS.get(metadata.type, "misc"),
            "ris": CM_TO_RIS_TRANSLATIONS.get(metadata.type, "GEN"),
        }
    )
    publication_year = (
        metadata.date.get("published")[:4]
        if metadata.date.get("published", None)
        else None
    )

    def to_datacite_date(date: dict) -> dict:
        """Convert dates to datacite dates"""
        for k, v in date.items():
            if k == "published":
                k = "issued"
            return {
                "date": v,
                "dateType": k.title(),
            }

    dates = [to_datacite_date(i) for i in wrap(metadata.date)]

    license_ = (
        [
            compact(
                {
                    "rightsIdentifier": metadata.license.get("id").lower()
                    if metadata.license.get("id", None)
                    else None,
                    "rightsIdentifierScheme": "SPDX",
                    "rightsUri": metadata.license.get("url", None),
                    "schemeUri": "https://spdx.org/licenses/",
                }
            )
        ]
        if metadata.license
        else None
    )

    data = compact(
        {
            "id": metadata.id,
            "doi": doi_from_url(metadata.id),
            "url": metadata.url,
            "creators": creators,
            "titles": metadata.titles,
            "publisher": metadata.publisher,
            "publicationYear": publication_year,
            "subjects": metadata.subjects,
            "contributors": contributors,
            "dates": dates,
            "language": metadata.language,
            "types": types,
            "relatedIdentifiers": related_identifiers,
            "sizes": metadata.sizes,
            "formats": metadata.formats,
            "version": metadata.version,
            "rightsList": license_,
            "descriptions": metadata.descriptions,
            "geoLocations": metadata.geo_locations,
            "fundingReferences": metadata.funding_references,
            "schemaVersion": "http://datacite.org/schema/kernel-4",
        }
    )
    return json.dumps(data)


def to_datacite_creator(creator: dict) -> dict:
    """Convert creators to datacite creators"""
    _type = creator.get("type", None)
    if creator.get("familyName", None):
        name = ", ".join([creator.get("familyName", ""), creator.get("givenName", "")])
    elif creator.get("name", None):
        name = creator.get("name", None)
    name_identifiers = creator.get("id", None)
    if name_identifiers:

        def format_name_identifier(name_identifier):
            return {
                "nameIdentifier": name_identifier,
                "nameIdentifierScheme": "ORCID",
                "schemeUri": "https://orcid.org",
            }

        name_identifiers = [format_name_identifier(i) for i in wrap(name_identifiers)]
    return compact(
        {
            "name": name,
            "givenName": creator.get("givenName", None),
            "familyName": creator.get("familyName", None),
            "nameType": _type + "al" if _type else None,
            "nameIdentifiers": name_identifiers,
            "affiliation": creator.get("affiliation", None),
        }
    )


def to_datacite_titles(titles: list) -> list:
    """Convert titles to datacite titles"""
    return [
        {
            "title": title.get("title", None),
            "titleType": title.get("type", None),
            "lang": title.get("language", None),
        }
        for title in titles
    ]


def to_datacite_related_identifier(reference: dict) -> dict:
    """Convert reference to datacite related_identifier"""
    doi = reference.get("doi", None)
    url = reference.get("url", None)
    return compact(
        {
            "relatedIdentifier": doi if doi else url,
            "relatedIdentifierType": "DOI" if doi else "URL",
            "relationType": "References",
        }
    )



"""Schema utils for commonmeta-py"""
from os import path
import orjson as json
from jsonschema import Draft202012Validator, ValidationError


def json_schema_errors(instance, schema: str = "commonmeta"):
    """validate against JSON schema"""
    schema_map = {
        "commonmeta": "commonmeta_v0.11",
        "datacite": "datacite-v4.5",
        "crossref": "crossref-v0.2",
        "csl": "csl-data",
        "cff": "cff_v1.2.0",
    }
    try:
        if schema not in schema_map.keys():
            raise ValueError("No schema found")
        file_path = path.join(
            path.dirname(__file__), f"resources/{schema_map[schema]}.json"
        )
        with open(file_path, encoding="utf-8") as file:
            string = file.read()
            schema = json.loads(string)
        return Draft202012Validator(schema).validate(instance)
    except ValidationError as error:
        return error.message



"""CSL-JSON reader for commonmeta-py"""
from ..utils import dict_to_spdx, from_csl, normalize_id, name_to_fos, encode_doi
from ..base_utils import wrap, compact, sanitize, presence
from ..author_utils import get_authors
from ..date_utils import get_date_from_date_parts
from ..doi_utils import get_doi_ra, doi_from_url
from ..constants import (
    CSL_TO_CM_TRANSLATIONS,
    Commonmeta,
)


def read_csl(data: dict, **kwargs) -> Commonmeta:
    """read_csl"""
    if data is None:
        return {"state": "not_found"}
    meta = data

    read_options = kwargs or {}

    _id = normalize_id(meta.get("id", None) or meta.get("DOI", None))
    _type = CSL_TO_CM_TRANSLATIONS.get(meta.get("type", None), "Other")

    # optionally generate a DOI if missing but a DOI prefix is provided
    prefix = read_options.get("prefix", None)
    if doi_from_url(_id) is None and prefix is not None:
        _id = encode_doi(prefix)

    contributors = get_authors(from_csl(wrap(meta.get("author", None))))
    contrib = get_authors(from_csl(wrap(meta.get("editor", None))))
    if contrib:
        contributors += contrib

    date = {"published": get_date_from_date_parts(meta.get("issued", None))}

    license_ = meta.get("copyright", None)
    if license_ is not None:
        license_ = dict_to_spdx({"url": meta.get("copyright")})

    pages = meta.get("page", "").split("-")
    publisher = meta.get("publisher", None)
    if isinstance(publisher, str):
        publisher = {"name": publisher}
    container = compact(
        {
            "type": "Periodical",
            "title": meta.get("container-title", None),
            "identifier": meta.get("ISSN", None),
            "identifierType": "ISSN" if meta.get("ISSN", None) else None,
            "volume": meta.get("volume", None),
            "issue": meta.get("issue", None),
            "firstPage": pages[0],
            "lastPage": pages[1] if len(pages) > 1 else None,
        }
    )

    state = "findable" if _id or read_options else "not_found"
    subjects = [name_to_fos(i) for i in wrap(meta.get("keywords", None))]

    if meta.get("abstract", None):
        descriptions = [
            {
                "description": sanitize(str(meta.get("abstract"))),
                "descriptionType": "Abstract",
            }
        ]
    else:
        descriptions = None

    provider = get_doi_ra(_id)

    return {
        "id": _id,
        "type": _type,
        "url": normalize_id(meta.get("URL", None)),
        "titles": [{"title": meta.get("title", None)}],
        "contributors": presence(contributors),
        "publisher": presence(publisher),
        "date": compact(date),
        "container": container,
        "references": None,
        "descriptions": descriptions,
        "license": license_,
        "version": meta.get("version", None),
        "subjects": subjects,
        "provider": provider,
        "state": state,
    } | read_options



"""Metadata"""
from os import path
import orjson as json
from typing import Optional, Union
import yaml
from pydash import py_

from .readers.crossref_reader import (
    get_crossref,
    read_crossref,
)
from .readers.datacite_reader import (
    get_datacite,
    read_datacite,
)
from .readers.datacite_xml_reader import read_datacite_xml
from .readers.crossref_xml_reader import (
    get_crossref_xml,
    read_crossref_xml,
)
from .readers.schema_org_reader import (
    get_schema_org,
    read_schema_org,
)
from .readers.codemeta_reader import (
    get_codemeta,
    read_codemeta,
)
from .readers.csl_reader import read_csl
from .readers.cff_reader import get_cff, read_cff
from .readers.json_feed_reader import get_json_feed_item, read_json_feed_item
from .readers.inveniordm_reader import (
    get_inveniordm,
    read_inveniordm,
)
from .readers.kbase_reader import read_kbase
from .readers.commonmeta_reader import read_commonmeta
from .readers.ris_reader import read_ris
from .writers.datacite_writer import write_datacite
from .writers.bibtex_writer import write_bibtex, write_bibtex_list
from .writers.citation_writer import write_citation, write_citation_list
from .writers.crossref_xml_writer import write_crossref_xml, write_crossref_xml_list
from .writers.csl_writer import write_csl, write_csl_list
from .writers.ris_writer import write_ris, write_ris_list
from .writers.schema_org_writer import write_schema_org
from .writers.commonmeta_writer import write_commonmeta, write_commonmeta_list
from .utils import normalize_id, find_from_format
from .base_utils import parse_xml, wrap
from .doi_utils import doi_from_url
from .schema_utils import json_schema_errors
from .constants import CM_TO_CR_TRANSLATIONS


# pylint: disable=R0902
class Metadata:
    """Metadata"""

    def __init__(self, string: Optional[Union[str, dict]], **kwargs):
        if string is None or not isinstance(string, (str, dict)):
            raise ValueError("No input found")
        self.via = kwargs.get("via", None)
        if isinstance(string, dict):
            data = string
        elif isinstance(string, str):
            pid = normalize_id(string)
            if pid is not None and self.via is None:
                self.via = find_from_format(pid=pid)
            elif path.exists(string):
                with open(string, encoding="utf-8") as file:
                    string = file.read()
                if self.via is None:
                    self.via = find_from_format(string=string)
            if self.via is None:
                self.via = "commonmeta"
            data = self.get_metadata(pid=pid, string=string)
        meta = self.read_metadata(data=data, **kwargs)

        # required properties
        self.id = meta.get("id")  # pylint: disable=C0103
        self.type = meta.get("type")
        # self.doi = meta.get("doi")
        self.url = meta.get("url")
        self.contributors = meta.get("contributors")
        self.titles = meta.get("titles")
        self.publisher = meta.get("publisher")
        self.date = meta.get("date")
        # recommended and optional properties
        self.additional_type = meta.get("additional_type")
        self.subjects = meta.get("subjects")
        self.language = meta.get("language")
        self.alternate_identifiers = meta.get("alternate_identifiers")
        self.related_identifiers = meta.get("related_identifiers")
        self.sizes = meta.get("sizes")
        self.formats = meta.get("formats")
        self.version = meta.get("version")
        self.license = meta.get("license")
        self.descriptions = meta.get("descriptions")
        self.geo_locations = meta.get("geo_locations")
        self.funding_references = meta.get("funding_references")
        self.references = meta.get("references")
        # other properties
        self.date_created = meta.get("date_created")
        self.date_registered = meta.get("date_registered")
        self.date_published = meta.get("date_published")
        self.date_updated = meta.get("date_updated")
        self.files = meta.get("files")
        self.container = meta.get("container")
        self.provider = meta.get("provider")
        self.schema_version = meta.get("schema_version")
        self.archive_locations = meta.get("archive_locations", None)

        # Catch errors in the reader, then validate against JSON schema for Commonmeta
        self.errors = meta.get("errors", None) or json_schema_errors(
            json.loads(self.write())
        )
        self.write_errors = None
        self.is_valid = (
            meta.get("state", None) != "not_found"
            and self.errors is None
            and self.write_errors is None
        )

    def get_metadata(self, pid, string) -> dict:
        via = self.via
        if pid is not None:
            if via == "schema_org":
                return get_schema_org(pid)
            elif via == "datacite":
                return get_datacite(pid)
            elif via in ["crossref", "op"]:
                return get_crossref(pid)
            elif via == "crossref_xml":
                return get_crossref_xml(pid)
            elif via == "codemeta":
                return get_codemeta(pid)
            elif via == "cff":
                return get_cff(pid)
            elif via == "json_feed_item":
                return get_json_feed_item(pid)
            elif via == "inveniordm":
                return get_inveniordm(pid)
        elif string is not None:
            if via == "datacite_xml":
                return parse_xml(string)
            elif via == "crossref_xml":
                return parse_xml(string, dialect="crossref")
            elif via == "cff":
                return yaml.safe_load(string)
            elif via == "bibtex":
                raise ValueError("Bibtex not supported")
            elif via == "ris":
                return string
            elif via in [
                "commonmeta",
                "crossref",
                "datacite",
                "schema_org",
                "csl",
                "json_feed_item",
                "codemeta",
                "kbase",
                "inveniordm",
            ]:
                return json.loads(string)
            else:
                raise ValueError("No input format found")
        else:
            raise ValueError("No metadata found")

    def read_metadata(self, data: dict, **kwargs) -> dict:
        """get_metadata"""
        via = self.via
        if via == "commonmeta":
            return read_commonmeta(data, **kwargs)
        elif via == "schema_org":
            return read_schema_org(data)
        elif via == "datacite":
            return read_datacite(data)
        elif via == "datacite_xml":
            return read_datacite_xml(data)
        elif via in ["crossref", "op"]:
            return read_crossref(data)
        elif via == "crossref_xml":
            return read_crossref_xml(data)
        elif via == "csl":
            return read_csl(data, **kwargs)
        elif via == "codemeta":
            return read_codemeta(data)
        elif via == "cff":
            return read_cff(data)
        elif via == "json_feed_item":
            return read_json_feed_item(data, **kwargs)
        elif via == "inveniordm":
            return read_inveniordm(data)
        elif via == "kbase":
            return read_kbase(data)
        elif via == "ris":
            return read_ris(data)
        else:
            raise ValueError("No input format found")

    def write(self, to: str = "commonmeta", **kwargs) -> str:
        """convert metadata into different formats"""
        try:
            if to == "commonmeta":
                return write_commonmeta(self)
            elif to == "bibtex":
                return write_bibtex(self)
            elif to == "csl":
                instance = py_.omit(json.loads(write_csl(self)), [])
                self.errors = json_schema_errors(instance, schema="csl")
                return write_csl(self)
            elif to == "citation":
                self.style = kwargs.get("style", "apa")
                self.locale = kwargs.get("locale", "en-US")
                return write_citation(self)
            elif to == "ris":
                return write_ris(self)
            elif to == "schema_org":
                return write_schema_org(self)
            elif to == "datacite":
                instance = json.loads(write_datacite(self))
                self.write_errors = json_schema_errors(instance, schema="datacite")
                return write_datacite(self)
            elif to == "crossref_xml":
                doi = doi_from_url(self.id)
                _type = CM_TO_CR_TRANSLATIONS.get(self.type, None)
                url = self.url
                instance = {"doi": doi, "type": _type, "url": url}
                self.depositor = kwargs.get("depositor", None)
                self.email = kwargs.get("email", None)
                self.registrant = kwargs.get("registrant", None)
                self.write_errors = json_schema_errors(instance, schema="crossref")
                return write_crossref_xml(self)
            else:
                raise ValueError("No output format found")
        except json.JSONDecodeError:
            raise ValueError("Invalid JSON")

    # legacy methods, to be removed in version 0.15
    def commonmeta(self):
        """Commonmeta"""
        return write_commonmeta(self)

    def bibtex(self):
        """Bibtex"""
        return write_bibtex(self)

    def csl(self):
        """CSL-JSON"""
        return write_csl(self)

    def citation(self):
        """Citation"""
        return write_citation(self)

    def ris(self):
        """RIS"""
        return write_ris(self)

    def schema_org(self):
        """Schema.org"""
        return write_schema_org(self)

    def datacite(self):
        """Datacite"""
        return write_datacite(self)

    def crossref_xml(self):
        """Crossref XML"""
        return write_crossref_xml(self)


class MetadataList:
    """MetadataList"""

    def __init__(
        self, dct: Optional[Union[str, dict]] = None, **kwargs
    ) -> Optional[dict]:
        if dct is None or not isinstance(dct, (str, bytes, dict)):
            raise ValueError("No input found")
        if isinstance(dct, dict):
            meta = dct
        elif isinstance(dct, (str, bytes)):
            if path.exists(dct):
                with open(dct, encoding="utf-8") as file:
                    dct = file.read()
            self.via = kwargs.get("via", None) or find_from_format(string=dct)
            meta = self.get_metadata_list(dct)

        self.id = meta.get("id", None)
        self.type = meta.get("type", None)
        self.title = meta.get("title", None)
        self.description = meta.get("description", None)

        # options needed for Crossref DOI registration
        self.depositor = kwargs.get("depositor", None)
        self.email = kwargs.get("email", None)
        self.registrant = kwargs.get("registrant", None)

        self.items = self.read_metadata_list(wrap(meta.get("items", None)), **kwargs)
        self.errors = [i.errors for i in self.items if i.errors is not None]
        self.write_errors = [
            i.write_errors for i in self.items if i.write_errors is not None
        ]
        self.is_valid = all([i.is_valid for i in self.items])

        # other options
        self.jsonlines = kwargs.get("jsonlines", False)
        self.filename = kwargs.get("filename", None)

    def get_metadata_list(self, string) -> list:
        if string is None or not isinstance(string, (str, bytes)):
            raise ValueError("No input found")
        if self.via in [
            "commonmeta",
            "crossref",
            "datacite",
            "schema_org",
            "csl",
            "json_feed_item",
        ]:
            return json.loads(string)
        else:
            raise ValueError("No input format found")

    def read_metadata_list(self, data: list, **kwargs) -> list:
        """read_metadata_list"""
        kwargs["via"] = kwargs.get("via", None) or self.via
        return [Metadata(i, **kwargs) for i in data]

    def write(self, to: str = "commonmeta", **kwargs) -> str:
        """convert metadata list into different formats"""
        if to == "commonmeta":
            return write_commonmeta_list(self)
        elif to == "bibtex":
            return write_bibtex_list(self)
        elif to == "csl":
            return write_csl_list(self)
        elif to == "citation":
            return write_citation_list(self, **kwargs)
        elif to == "ris":
            return write_ris_list(self)
        elif to == "schema_org":
            raise ValueError("Schema.org not supported for metadata lists")
        elif to == "datacite":
            raise ValueError("Datacite not supported for metadata lists")
        elif to == "crossref_xml":
            return write_crossref_xml_list(self)
        else:
            raise ValueError("No output format found")



# -*- coding: utf-8 -*-

# commonmeta-py

"""
commonmeta-py library
~~~~~~~~~~~~~~~~~~~~~

commonmeta-py is a Python library to convert scholarly metadata
"""

__title__ = "commonmeta-py"
__version__ = "0.13.0"
__author__ = "Martin Fenner"
__license__ = "MIT"

# ruff: noqa: F401
from .metadata import Metadata, MetadataList
from .readers import (
    cff_reader,
    codemeta_reader,
    crossref_reader,
    crossref_xml_reader,
    datacite_reader,
    datacite_xml_reader,
    inveniordm_reader,
    json_feed_reader,
    kbase_reader,
    ris_reader,
    schema_org_reader,
)
from .writers import (
    bibtex_writer,
    citation_writer,
    commonmeta_writer,
    csl_writer,
    datacite_writer,
    ris_writer,
    schema_org_writer,
)
from .utils import (
    dict_to_spdx,
    from_csl,
    from_schema_org,
    get_language,
    normalize_cc_url,
    normalize_id,
    normalize_ids,
    normalize_orcid,
    normalize_url,
    normalize_ror,
    pages_as_string,
    to_csl,
    validate_orcid,
    validate_url,
    get_language,
    encode_doi,
    name_to_fos,
    from_json_feed,
)
from .author_utils import (
    authors_as_string,
    cleanup_author,
    get_affiliations,
    get_authors,
    get_one_author,
    is_personal_name,
)
from .base_utils import (
    wrap,
    unwrap,
    compact,
    presence,
    parse_attributes,
    sanitize,
)
from .date_utils import (
    get_date_from_crossref_parts,
    get_date_from_date_parts,
    get_date_from_unix_timestamp,
    get_date_parts,
    get_iso8601_date,
    strip_milliseconds,
)
from .doi_utils import (
    crossref_api_url,
    crossref_xml_api_url,
    doi_from_url,
    doi_as_url,
    doi_resolver,
    datacite_api_url,
    get_doi_ra,
    normalize_doi,
    validate_doi,
    validate_prefix,
)



"""Utils module for commonmeta-py"""
import os
import orjson as json
import re
import time
from typing import Optional
from urllib.parse import urlparse
import yaml
from furl import furl
import bibtexparser
from bs4 import BeautifulSoup
from pydash import py_
import base32_lib as base32
import pycountry

from .base_utils import wrap, compact, parse_attributes
from .doi_utils import normalize_doi, doi_from_url, get_doi_ra, validate_doi, doi_as_url
from .constants import DATACITE_CONTRIBUTOR_TYPES


NORMALIZED_LICENSES = {
    "https://creativecommons.org/licenses/by/1.0": "https://creativecommons.org/licenses/by/1.0/legalcode",
    "https://creativecommons.org/licenses/by/2.0": "https://creativecommons.org/licenses/by/2.0/legalcode",
    "https://creativecommons.org/licenses/by/2.5": "https://creativecommons.org/licenses/by/2.5/legalcode",
    "https://creativecommons.org/licenses/by/3.0": "https://creativecommons.org/licenses/by/3.0/legalcode",
    "https://creativecommons.org/licenses/by/3.0/us": "https://creativecommons.org/licenses/by/3.0/legalcode",
    "https://creativecommons.org/licenses/by/4.0": "https://creativecommons.org/licenses/by/4.0/legalcode",
    "https://creativecommons.org/licenses/by-nc/1.0": "https://creativecommons.org/licenses/by-nc/1.0/legalcode",
    "https://creativecommons.org/licenses/by-nc/2.0": "https://creativecommons.org/licenses/by-nc/2.0/legalcode",
    "https://creativecommons.org/licenses/by-nc/2.5": "https://creativecommons.org/licenses/by-nc/2.5/legalcode",
    "https://creativecommons.org/licenses/by-nc/3.0": "https://creativecommons.org/licenses/by-nc/3.0/legalcode",
    "https://creativecommons.org/licenses/by-nc/4.0": "https://creativecommons.org/licenses/by-nc/4.0/legalcode",
    "https://creativecommons.org/licenses/by-nd-nc/1.0": "https://creativecommons.org/licenses/by-nd-nc/1.0/legalcode",
    "https://creativecommons.org/licenses/by-nd-nc/2.0": "https://creativecommons.org/licenses/by-nd-nc/2.0/legalcode",
    "https://creativecommons.org/licenses/by-nd-nc/2.5": "https://creativecommons.org/licenses/by-nd-nc/2.5/legalcode",
    "https://creativecommons.org/licenses/by-nd-nc/3.0": "https://creativecommons.org/licenses/by-nd-nc/3.0/legalcode",
    "https://creativecommons.org/licenses/by-nd-nc/4.0": "https://creativecommons.org/licenses/by-nd-nc/4.0/legalcode",
    "https://creativecommons.org/licenses/by-nc-sa/1.0": "https://creativecommons.org/licenses/by-nc-sa/1.0/legalcode",
    "https://creativecommons.org/licenses/by-nc-sa/2.0": "https://creativecommons.org/licenses/by-nc-sa/2.0/legalcode",
    "https://creativecommons.org/licenses/by-nc-sa/2.5": "https://creativecommons.org/licenses/by-nc-sa/2.5/legalcode",
    "https://creativecommons.org/licenses/by-nc-sa/3.0": "https://creativecommons.org/licenses/by-nc-sa/3.0/legalcode",
    "https://creativecommons.org/licenses/by-nc-sa/3.0/us": "https://creativecommons.org/licenses/by-nc-sa/3.0/legalcode",
    "https://creativecommons.org/licenses/by-nc-sa/4.0": "https://creativecommons.org/licenses/by-nc-sa/4.0/legalcode",
    "https://creativecommons.org/licenses/by-nd/1.0": "https://creativecommons.org/licenses/by-nd/1.0/legalcode",
    "https://creativecommons.org/licenses/by-nd/2.0": "https://creativecommons.org/licenses/by-nd/2.0/legalcode",
    "https://creativecommons.org/licenses/by-nd/2.5": "https://creativecommons.org/licenses/by-nd/2.5/legalcode",
    "https://creativecommons.org/licenses/by-nd/3.0": "https://creativecommons.org/licenses/by-nd/3.0/legalcode",
    "https://creativecommons.org/licenses/by-nd/4.0": "https://creativecommons.org/licenses/by-nd/2.0/legalcode",
    "https://creativecommons.org/licenses/by-sa/1.0": "https://creativecommons.org/licenses/by-sa/1.0/legalcode",
    "https://creativecommons.org/licenses/by-sa/2.0": "https://creativecommons.org/licenses/by-sa/2.0/legalcode",
    "https://creativecommons.org/licenses/by-sa/2.5": "https://creativecommons.org/licenses/by-sa/2.5/legalcode",
    "https://creativecommons.org/licenses/by-sa/3.0": "https://creativecommons.org/licenses/by-sa/3.0/legalcode",
    "https://creativecommons.org/licenses/by-sa/4.0": "https://creativecommons.org/licenses/by-sa/4.0/legalcode",
    "https://creativecommons.org/licenses/by-nc-nd/1.0": "https://creativecommons.org/licenses/by-nc-nd/1.0/legalcode",
    "https://creativecommons.org/licenses/by-nc-nd/2.0": "https://creativecommons.org/licenses/by-nc-nd/2.0/legalcode",
    "https://creativecommons.org/licenses/by-nc-nd/2.5": "https://creativecommons.org/licenses/by-nc-nd/2.5/legalcode",
    "https://creativecommons.org/licenses/by-nc-nd/3.0": "https://creativecommons.org/licenses/by-nc-nd/3.0/legalcode",
    "https://creativecommons.org/licenses/by-nc-nd/4.0": "https://creativecommons.org/licenses/by-nc-nd/4.0/legalcode",
    "https://creativecommons.org/licenses/publicdomain": "https://creativecommons.org/licenses/publicdomain/",
    "https://creativecommons.org/publicdomain/zero/1.0": "https://creativecommons.org/publicdomain/zero/1.0/legalcode",
}

UNKNOWN_INFORMATION = {
    ":unac": "temporarily inaccessible",
    ":unal": "unallowed, suppressed intentionally",
    ":unap": "not applicable, makes no sense",
    ":unas": "value unassigned (e.g., Untitled)",
    ":unav": "value unavailable, possibly unknown",
    ":unkn": "known to be unknown (e.g., Anonymous, Inconnue)",
    ":none": "never had a value, never will",
    ":null": "explicitly and meaningfully empty",
    ":tba": "to be assigned or announced later",
    ":etal": "too numerous to list (et alia)",
}

HTTP_SCHEME = "http://"
HTTPS_SCHEME = "https://"


def normalize_id(pid: Optional[str], **kwargs) -> Optional[str]:
    """Check for valid DOI or HTTP(S) URL"""
    if pid is None:
        return None

    # check if pid is a bytes object
    if isinstance(pid, (bytes, bytearray)):
        pid = pid.decode()

    # check for valid DOI
    doi = normalize_doi(pid, **kwargs)
    if doi is not None:
        return doi

    # check for valid HTTP uri and ensure https
    uri = urlparse(pid)
    if not uri.netloc or uri.scheme not in ["http", "https"]:
        return None
    if uri.scheme == "http":
        pid = pid.replace(HTTP_SCHEME, HTTPS_SCHEME)

    # remove trailing slash
    if pid.endswith("/"):
        pid = pid.strip("/")

    return pid


def normalize_ids(ids: list, relation_type=None) -> list:
    """Normalize identifiers"""

    def format_id(i):
        if i.get("id", None):
            idn = normalize_id(i["id"])
            doi = doi_from_url(idn)
            related_identifier_type = "DOI" if doi is not None else "URL"
            idn = doi or idn
            _type = (
                i.get("type")
                if isinstance(i.get("type", None), str)
                else wrap(i.get("type", None))[0]
            )
            return compact(
                {
                    "relatedIdentifier": idn,
                    "relationType": relation_type,
                    "relatedIdentifierType": related_identifier_type,
                }
            )
        return None

    return [format_id(i) for i in ids]


def normalize_url(url: Optional[str], secure=False, lower=False) -> Optional[str]:
    """Normalize URL"""
    if url is None or not isinstance(url, str):
        return None
    if url.endswith("/"):
        url = url.strip("/")
    if secure is True and url.startswith(HTTP_SCHEME):
        url = url.replace(HTTP_SCHEME, HTTPS_SCHEME)
    if lower is True:
        return url.lower()
    return url


def normalize_cc_url(url: Optional[str]):
    """Normalize Creative Commons URL"""
    if url is None or not isinstance(url, str):
        return None
    url = normalize_url(url, secure=True)
    return NORMALIZED_LICENSES.get(url, url)


def normalize_ror(ror: Optional[str]) -> Optional[str]:
    """Normalize ROR ID"""
    ror = validate_ror(ror)
    if ror is None:
        return None

    # turn ROR ID into URL
    return "https://ror.org/" + ror


def validate_ror(ror: Optional[str]) -> Optional[str]:
    """Validate ROR"""
    if ror is None or not isinstance(ror, str):
        return None
    match = re.search(
        r"\A(?:(?:http|https)://ror\.org/)?([0-9a-z]{7}\d{2})\Z",
        ror,
    )
    if match is None:
        return None
    ror = match.group(1).replace(" ", "-")
    return ror


def validate_url(url: str) -> Optional[str]:
    if url is None:
        return None
    elif validate_doi(url):
        return "DOI"
    f = furl(url)
    if f and f.scheme in ["http", "https"]:
        return "URL"
    match = re.search(
        r"\A(ISSN|eISSN) (\d{4}-\d{3}[0-9X]+)\Z",
        url,
    )
    if match is not None:
        return "ISSN"
    return None


def normalize_orcid(orcid: Optional[str]) -> Optional[str]:
    """Normalize ORCID"""
    if orcid is None or not isinstance(orcid, str):
        return None
    orcid = validate_orcid(orcid)
    if orcid is None:
        return None
    return "https://orcid.org/" + orcid


def validate_orcid(orcid: Optional[str]) -> Optional[str]:
    """Validate ORCID"""
    if orcid is None or not isinstance(orcid, str):
        return None
    match = re.search(
        r"\A(?:(?:http|https)://(?:(?:www|sandbox)?\.)?orcid\.org/)?(\d{4}[ -]\d{4}[ -]\d{4}[ -]\d{3}[0-9X]+)\Z",
        orcid,
    )
    if match is None:
        return None
    orcid = match.group(1).replace(" ", "-")
    return orcid


def validate_isni(isni: Optional[str]) -> Optional[str]:
    """Validate ISNI"""
    if isni is None or not isinstance(isni, str):
        return None
    match = re.search(
        r"\A(?:(?:http|https)://isni\.org/isni/)?(\d{4}([ -])?\d{4}([ -])?\d{4}([ -])?\d{3}[0-9X]+)\Z",
        isni,
    )
    if match is None:
        return None
    isni = match.group(1).replace(" ", "")
    return isni


def normalize_isni(isni: Optional[str]) -> Optional[str]:
    """Normalize ISNI"""
    if isni is None or not isinstance(isni, str):
        return None
    isni = validate_isni(isni)
    if isni is None:
        return None
    return "https://isni.org/isni/" + isni


def normalize_name_identifier(ni: Optional[str]) -> Optional[str]:
    """Normalize name identifier"""
    if ni is None:
        return None
    if isinstance(ni, str):
        return
    if isinstance(ni, dict):
        return format_name_identifier(ni)
    if isinstance(ni, list):
        return next(
            (format_name_identifier(i) for i in wrap(ni.get("nameIdentifiers", None))),
            None,
        )
    return None


def format_name_identifier(ni):
    """format_name_identifier"""
    if ni is None:
        return None
    elif isinstance(ni, str):
        return normalize_orcid(ni) or normalize_ror(ni) or normalize_isni(ni)
    name_identifier = ni.get("nameIdentifier", None) or ni.get(
        "publisherIdentifier", None
    )
    name_identifier_scheme = ni.get("nameIdentifierScheme", None) or ni.get(
        "publisherIdentifierScheme", None
    )
    scheme_uri = ni.get("schemeURI", None) or ni.get("schemeUri", None)
    if name_identifier is None:
        return None
    elif name_identifier_scheme == "ORCID":
        return normalize_orcid(name_identifier)
    elif name_identifier_scheme == "ISNI":
        return normalize_isni(name_identifier)
    elif name_identifier_scheme == "ROR":
        return normalize_ror(name_identifier)
    elif validate_url(name_identifier) == "URL":
        return name_identifier
    elif isinstance(name_identifier, str) and scheme_uri is not None:
        return scheme_uri + name_identifier
    return None


def normalize_issn(string, **kwargs):
    """Normalize ISSN
    Pick electronic issn if there are multiple
    Format issn as xxxx-xxxx"""
    content = kwargs.get("content", "#text")
    if string is None:
        return None
    if isinstance(string, str):
        issn = string
    elif isinstance(string, dict):
        issn = string.get(content, None)
    elif isinstance(string, list):
        issn = next(
            (i for i in string if i.get("media_type", None) == "electronic"), {}
        ).get(content, None)
    if issn is None:
        return None
    if len(issn) == 9:
        return issn
    if len(issn) == 8:
        return issn[0:4] + "-" + issn[4:8]
    return None


def dict_to_spdx(dct: dict) -> dict:
    """Convert a dict to SPDX"""
    dct.update({"url": normalize_cc_url(dct.get("url", None))})
    file_path = os.path.join(
        os.path.dirname(__file__), "resources", "spdx", "licenses.json"
    )
    with open(file_path, encoding="utf-8") as file:
        string = file.read()
        spdx = json.loads(string).get("licenses")
    license_ = next(
        (
            lic
            for lic in spdx
            if lic["licenseId"].casefold() == dct.get("id", "").casefold()
            or lic["seeAlso"][0] == dct.get("url", None)
        ),
        None,
    )
    if license_ is None:
        return compact(dct)
    #   license = spdx.find do |l|
    #     l['licenseId'].casecmp?(hsh['rightsIdentifier']) || l['seeAlso'].first == normalize_cc_url(hsh['rightsUri']) || l['name'] == hsh['rights'] || l['seeAlso'].first == normalize_cc_url(hsh['rights'])
    #   end
    return compact(
        {
            "id": license_["licenseId"],
            "url": license_["seeAlso"][0],
        }
    )

    #   else
    #     {
    #       'rights': hsh['#text'] || hsh['rights'],
    #       'rightsUri': hsh['rightsUri'] || hsh['rightsUri'],
    #       'rightsIdentifier': hsh['rightsIdentifier'].present? ? hsh['rightsIdentifier'].downcase : None,
    #       'rightsIdentifierScheme': hsh['rightsIdentifierScheme'],
    #       'schemeUri': hsh['schemeUri'],
    #       'lang': hsh['lang']
    #     }.compact
    #   end
    # end


def from_json_feed(elements: list) -> list:
    """Convert from JSON Feed elements"""

    def format_element(element):
        """format element"""
        if not isinstance(element, dict):
            return None
        mapping = {"url": "id"}
        for key, value in mapping.items():
            if element.get(key, None) is not None:
                element[value] = element.pop(key)
        return element

    return [format_element(i) for i in elements]


def from_inveniordm(elements: list) -> list:
    """Convert from inveniordm elements"""

    def format_element(element):
        """format element"""
        if not isinstance(element, dict):
            return None
        mapping = {"orcid": "ORCID"}
        for key, value in mapping.items():
            if element.get(key, None) is not None:
                element[value] = element.pop(key)
        return element

    return [format_element(i) for i in elements]


def from_crossref_xml(elements: list) -> list:
    """Convert from crossref_xml elements"""

    def format_affiliation(element):
        """Format affiliation"""
        return {"name": element}

    def format_element(element):
        """format element"""
        if element.get("name", None) is not None:
            element["type"] = "Organization"
            element["name"] = element.get("name")
        else:
            element["type"] = "Person"
        element["givenName"] = element.get("given_name", None)
        element["familyName"] = element.get("surname", None)
        element["contributorType"] = element.get(
            "contributor_role", "author"
        ).capitalize()
        if element.get("ORCID", None) is not None:
            orcid = parse_attributes(element.get("ORCID"))
            element["ORCID"] = normalize_orcid(orcid)
        element = py_.omit(
            element, "given_name", "surname", "sequence", "contributor_role"
        )
        return compact(element)

    return [format_element(i) for i in elements]


def from_kbase(elements: list) -> list:
    """Convert from kbase elements"""

    def map_contributor_role(role):
        if role.split(":")[0] == "CRediT":
            return py_.pascal_case(role.split(":")[1])
        elif role.split(":")[0] == "DataCite":
            return DATACITE_CONTRIBUTOR_TYPES.get(role.split(":")[1], "Other")
        else:
            return role.split(":")[1]

    def format_element(element):
        """format element"""
        if not isinstance(element, dict):
            return None
        if element.get("contributor_id", None) is not None:
            element["ORCID"] = from_curie(element["contributor_id"])
        element["contributor_roles"] = [
            map_contributor_role(i)
            for i in wrap(element.get("contributor_roles", None))
        ]
        element = py_.omit(element, "contributor_id")
        return compact(element)

    return [format_element(i) for i in elements]


def from_csl(elements: list) -> list:
    """Convert from csl elements"""

    def format_element(element):
        """format element"""
        if element.get("literal", None) is not None:
            element["type"] = "Organization"
            element["name"] = element["literal"]
        elif element.get("name", None) is not None:
            element["type"] = "Organization"
            element["name"] = element.get("name")
        else:
            element["type"] = "Person"
            element["name"] = " ".join(
                [element.get("given", ""), element.get("family", "")]
            )
        element["givenName"] = element.get("given", None)
        element["familyName"] = element.get("family", None)
        element["affiliation"] = element.get("affiliation", None)
        element = py_.omit(element, "given", "family", "literal", "sequence")
        return compact(element)

    return [format_element(i) for i in elements]


def to_csl(elements: list) -> list:
    """Convert elements to CSL-JSON"""

    def format_element(i):
        """format element"""
        element = {}
        element["family"] = i.get("familyName", None)
        element["given"] = i.get("givenName", None)
        element["literal"] = (
            i.get("name", None) if i.get("familyName", None) is None else None
        )
        return compact(element)

    return [format_element(i) for i in elements]


def to_ris(elements: Optional[list]) -> list:
    """Convert element to RIS"""
    if elements is None:
        return []

    def format_element(i):
        """format element"""
        if i.get("familyName", None) and i.get("givenName", None):
            element = ", ".join([i["familyName"], i.get("givenName", None)])
        else:
            element = i.get("name", None)
        return element

    return [
        format_element(i)
        for i in elements
        if i.get("name", None) or i.get("familyName", None)
    ]


def to_schema_org(element: Optional[dict]) -> Optional[dict]:
    """Convert a metadata element to Schema.org"""
    if not isinstance(element, dict):
        return None
    mapping = {"type": "@type", "id": "@id", "title": "name"}
    for key, value in mapping.items():
        if element.get(key, None) is not None:
            element[value] = element.pop(key)
    return element


def to_schema_org_creators(elements: list) -> list():
    """Convert creators to Schema.org"""

    def format_element(element):
        """format element"""
        element["@type"] = element["type"][0:-2] if element.get("type", None) else None
        if element.get("familyName", None) and element.get("name", None) is None:
            element["name"] = " ".join(
                [element.get("givenName", None), element.get("familyName")]
            )
            element["@type"] = "Person"
        else:
            element["@type"] = "Organization"
        element = py_.omit(element, "type", "contributorRoles")
        return compact(element)

    return [format_element(i) for i in elements]


def to_schema_org_container(element: Optional[dict], **kwargs) -> Optional[dict]:
    """Convert CSL container to Schema.org container"""
    if element is None and kwargs.get("container_title", None) is None:
        return None
    if not isinstance(element, dict):
        return None

    return compact(
        {
            "@id": element.get("identifier", None),
            "@type": "DataCatalog"
            if kwargs.get("type", None) == "DataRepository"
            else "Periodical",
            "name": element.get("title", None) or kwargs.get("container_title", None),
        }
    )


def to_schema_org_identifiers(elements: list) -> list:
    """Convert identifiers to Schema.org"""

    def format_element(i):
        """format element"""
        element = {}
        element["@type"] = "PropertyValue"
        element["propertyID"] = i.get("identifierType", None)
        element["value"] = i.get("identifier", None)
        return compact(element)

    return [format_element(i) for i in elements]


def to_schema_org_relations(related_items: list, relation_type=None):
    """Convert relatedItems to Schema.org relations"""

    def format_element(i):
        """format element"""
        if i["relatedItemIdentifierType"] == "ISSN" and i["relationType"] == "IsPartOf":
            return compact({"@type": "Periodical", "issn": i["relatedItemIdentifier"]})
        return compact({"@id": normalize_id(i["relatedIdentifier"])})

    # consolidate different relation types
    if relation_type == "References":
        relation_type = ["References", "Cites"]
    else:
        relation_type = [relation_type]

    related_items = py_.filter(
        wrap(related_items), lambda ri: ri["relationType"] in relation_type
    )
    return [format_element(i) for i in related_items]


def find_from_format(pid=None, string=None, ext=None, dct=None, filename=None):
    """Find reader from format"""
    if pid is not None:
        return find_from_format_by_id(pid)
    if string is not None and ext is not None:
        return find_from_format_by_ext(ext)
    if dct is not None:
        return find_from_format_by_dict(dct)
    if string is not None:
        return find_from_format_by_string(string)
    if filename is not None:
        return find_from_format_by_filename(filename)
    return "datacite"


def find_from_format_by_id(pid: str) -> Optional[str]:
    """Find reader from format by id"""
    doi = validate_doi(pid)
    if doi and (registration_agency := get_doi_ra(doi)) is not None:
        return registration_agency.lower()
    if (
        re.match(r"\A(http|https):/(/)?github\.com/(.+)/CITATION.cff\Z", pid)
        is not None
    ):
        return "cff"
    if (
        re.match(r"\A(http|https):/(/)?github\.com/(.+)/codemeta.json\Z", pid)
        is not None
    ):
        return "codemeta"
    if re.match(r"\A(http|https):/(/)?github\.com/(.+)\Z", pid) is not None:
        return "cff"
    if re.match(r"\Ahttps:/(/)?api\.rogue-scholar\.org/posts/(.+)\Z", pid) is not None:
        return "json_feed_item"
    if re.match(r"\Ahttps:/(/)?zenodo\.org/api/records/(.+)\Z", pid) is not None:
        return "inveniordm"
    return "schema_org"


def find_from_format_by_ext(ext: str) -> Optional[str]:
    """Find reader from format by ext"""
    if ext == ".bib":
        return "bibtex"
    if ext == ".ris":
        return "ris"
    return None


def find_from_format_by_dict(dct: dict) -> Optional[str]:
    if dct is None or not isinstance(dct, dict):
        return None
    """Find reader from format by dict"""
    if dct.get("schema_version", "").startswith("https://commonmeta.org"):
        return "commonmeta"
    if dct.get("@context", None) == "http://schema.org":
        return "schema_org"
    if dct.get("@context", None) in [
        "https://raw.githubusercontent.com/codemeta/codemeta/master/codemeta.jsonld"
    ]:
        return "codemeta"
    if py_.get(dct, "blog.version", None) == "https://jsonfeed.org/version/1.1":
        return "json_feed_item"
    if dct.get("schemaVersion", "").startswith("http://datacite.org/schema/kernel"):
        return "datacite"
    if dct.get("source", None) == "Crossref":
        return "crossref"
    if py_.get(dct, "issued.date-parts") is not None:
        return "csl"
    if py_.get(dct, "conceptdoi") is not None:
        return "inveniordm"
    if py_.get(dct, "credit_metadata") is not None:
        return "kbase"
    return None


def find_from_format_by_string(string: str) -> Optional[str]:
    """Find reader from format by string"""
    if string is None:
        return None
    try:
        data = json.loads(string)
        if not isinstance(data, dict):
            raise TypeError
        if data.get("items", None) is not None:
            data = data["items"][0]
        if data.get("schema_version", "").startswith("https://commonmeta.org"):
            return "commonmeta"
        if data.get("@context", None) == "http://schema.org":
            return "schema_org"
        if data.get("@context", None) in [
            "https://raw.githubusercontent.com/codemeta/codemeta/master/codemeta.jsonld"
        ]:
            return "codemeta"
        if py_.get(data, "blog.version", None) == "https://jsonfeed.org/version/1.1":
            return "json_feed_item"
        if data.get("schemaVersion", "").startswith(
            "http://datacite.org/schema/kernel"
        ):
            return "datacite"
        if data.get("source", None) == "Crossref":
            return "crossref"
        if py_.get(data, "issued.date-parts") is not None:
            return "csl"
        if py_.get(data, "conceptdoi") is not None:
            return "inveniordm"
        if py_.get(data, "credit_metadata") is not None:
            return "kbase"
    except (TypeError, json.JSONDecodeError):
        pass
    try:
        data = BeautifulSoup(string, "xml")
        if data.find("doi_record"):
            return "crossref_xml"
        if data.find("resource"):
            return "datacite_xml"
    except ValueError:
        pass
    try:
        data = yaml.safe_load(string)
        if data.get("cff-version", None):
            return "cff"
    except (yaml.YAMLError, AttributeError):
        pass

    if string.startswith("TY  - "):
        return "ris"
    if any(string.startswith(f"@{t}") for t in bibtexparser.bibdatabase.STANDARD_TYPES):
        return "bibtex"

    # no format found
    return None

    # if Maremma.from_xml(string).to_h.dig('crossref_result', 'query_result', 'body', 'query',
    #                                        'doi_record', 'crossref').present?
    #     'crossref_xml'
    #   elsif Nokogiri::XML(string, None, 'UTF-8', &:noblanks).collect_namespaces.find do |_k, v|
    #           v.start_with?('http://datacite.org/schema/kernel')
    # #         end
    #     'datacite_xml'


def find_from_format_by_filename(filename):
    """Find reader from format by filename"""
    if filename == "CITATION.cff":
        return "cff"
    return None


def from_schema_org(element):
    """Convert schema.org to DataCite"""
    if element is None:
        return None
    element["type"] = element.get("@type", None)
    element["id"] = element.get("@id", None)
    return compact(py_.omit(element, ["@type", "@id"]))


def from_schema_org_creators(elements: list) -> list:
    """Convert schema.org creators to commonmeta"""

    def format_element(i):
        """format element"""
        element = {}
        if urlparse(i.get("@id", None)).hostname == "orcid.org":
            element["id"] = i.get("@id")
            element["type"] = "Person"
        elif isinstance(i.get("@type", None), str):
            element["type"] = i.get("@type")
        elif isinstance(i.get("@type", None), list):
            element["type"] = py_.find(
                i["@type"], lambda x: x in ["Person", "Organization"]
            )

        # strip text after comma if suffix is an academic title
        if str(i["name"]).split(", ", maxsplit=1)[-1] in [
            "MD",
            "PhD",
            "DVM",
            "DDS",
            "DMD",
            "JD",
            "MBA",
            "MPH",
            "MS",
            "MA",
            "MFA",
            "MSc",
            "MEd",
            "MEng",
            "MPhil",
            "MRes",
            "LLM",
            "LLB",
            "BSc",
            "BA",
            "BFA",
            "BEd",
            "BEng",
            "BPhil",
        ]:
            i["name"] = str(i["name"]).split(", ", maxsplit=1)[0]
        length = len(str(i["name"]).split(" "))
        if i.get("givenName", None):
            element["givenName"] = i.get("givenName", None)
        if i.get("familyName", None):
            element["familyName"] = i.get("familyName", None)
            element["type"] = "Person"
        # parentheses around the last word indicate an organization
        elif length > 1 and not str(i["name"]).rsplit(" ", maxsplit=1)[-1].startswith(
            "("
        ):
            element["givenName"] = " ".join(str(i["name"]).split(" ")[0 : length - 1])
            element["familyName"] = str(i["name"]).rsplit(" ", maxsplit=1)[1:]
        if not element.get("familyName", None):
            element["creatorName"] = compact(
                {
                    "type": i.get("@type", None),
                    "#text": i.get("name", None),
                }
            )

        if isinstance(i.get("affiliation", None), str):
            element["affiliation"] = {"type": "Organization", "name": i["affiliation"]}
        elif urlparse(py_.get(i, "affiliation.@id", "")).hostname in [
            "ror.org",
            "isni.org",
        ]:
            element["affiliation"] = {
                "id": i["affiliation"]["@id"],
                "type": "Organization",
                "name": i["affiliation"]["name"],
            }
        return compact(element)

    return [format_element(i) for i in wrap(elements)]


def github_from_url(url: str) -> dict:
    """Get github owner, repo, release and path from url"""

    match = re.match(
        r"\Ahttps://(github|raw\.githubusercontent)\.com/(.+)(?:/)?(.+)?(?:/tree/)?(.*)\Z",
        url,
    )
    if match is None:
        return {}
    words = urlparse(url).path.lstrip("/").split("/")
    owner = words[0] if len(words) > 0 else None
    repo = words[1] if len(words) > 1 else None
    release = words[3] if len(words) > 3 else None
    path = "/".join(words[4:]) if len(words) > 3 else ""
    if len(path) == 0:
        path = None

    return compact({"owner": owner, "repo": repo, "release": release, "path": path})


def github_repo_from_url(url: str) -> Optional[str]:
    """Get github repo from url"""
    return github_from_url(url).get("repo", None)


def github_release_from_url(url: str) -> Optional[str]:
    """Get github release from url"""
    return github_from_url(url).get("release", None)


def github_owner_from_url(url: str) -> Optional[str]:
    """Get github owner from url"""
    return github_from_url(url).get("owner", None)


def github_as_owner_url(url: str) -> Optional[str]:
    """Get github owner url from url"""
    github_dict = github_from_url(url)
    if github_dict.get("owner", None) is None:
        return None
    return f"https://github.com/{github_dict.get('owner')}"


def github_as_repo_url(url) -> Optional[str]:
    """Get github repo url from url"""
    github_dict = github_from_url(url)
    if github_dict.get("repo", None) is None:
        return None
    return f"https://github.com/{github_dict.get('owner')}/{github_dict.get('repo')}"


def github_as_release_url(url: str) -> Optional[str]:
    """Get github release url from url"""
    github_dict = github_from_url(url)
    if github_dict.get("release", None) is None:
        return None
    return f"https://github.com/{github_dict.get('owner')}/{github_dict.get('repo')}/tree/{github_dict.get('release')}"


def github_as_codemeta_url(url: str) -> Optional[str]:
    """Get github codemeta.json url from url"""
    github_dict = github_from_url(url)

    if github_dict.get("path", None) and github_dict.get("path").endswith(
        "codemeta.json"
    ):
        return f"https://raw.githubusercontent.com/{github_dict.get('owner')}/{github_dict.get('repo')}/{github_dict.get('release')}/{github_dict.get('path')}"
    elif github_dict.get("owner", None):
        return f"https://raw.githubusercontent.com/{github_dict.get('owner')}/{github_dict.get('repo')}/master/codemeta.json"
    else:
        return None


def github_as_cff_url(url: str) -> Optional[str]:
    """Get github CITATION.cff url from url"""
    github_dict = github_from_url(url)

    if github_dict.get("path", None) and github_dict.get("path").endswith(
        "CITATION.cff"
    ):
        return f"https://raw.githubusercontent.com/{github_dict.get('owner')}/{github_dict.get('repo')}/{github_dict.get('release')}/{github_dict.get('path')}"
    if github_dict.get("owner", None):
        return f"https://raw.githubusercontent.com/{github_dict.get('owner')}/{github_dict.get('repo')}/main/CITATION.cff"
    return None


def pages_as_string(
    container: Optional[dict], page_range_separator="-"
) -> Optional[str]:
    """Parse pages for BibTeX"""
    if container is None:
        return None
    if container.get("firstPage", None) is None:
        return None
    if container.get("lastPage", None) is None:
        return container.get("firstPage", None)

    return page_range_separator.join(
        [container.get("firstPage"), container.get("lastPage", None)]
    )


def subjects_as_string(subjects):
    """convert subject list to string, e.g. for bibtex"""
    if subjects is None:
        return None

    keywords = []
    for subject in wrap(subjects):
        keywords.append(subject.get("subject", None))
    return ", ".join(keywords)


# def reverse():
#       return { 'citation': wrap(related_identifiers).select do |ri|
#                         ri['relationType'] == 'IsReferencedBy'
#                       end.map do |r|
#                         { '@id': normalize_doi(r['relatedIdentifier']),
#                           '@type': r['resourceTypeGeneral'] validate_orcid 'ScholarlyArticle',
#                           'identifier': r['relatedIdentifierType'] == 'DOI' ? nil : to_identifier(r) }.compact
#                       end.unwrap,
#         'isBasedOn': wrap(related_identifiers).select do |ri|
#                          ri['relationType'] == 'IsSupplementTo'
#                        end.map do |r|
#                          { '@id': normalize_doi(r['relatedIdentifier']),
#                            '@type': r['resourceTypeGeneral'] or 'ScholarlyArticle',
#                            'identifier': r['relatedIdentifierType'] == 'DOI' ? nil : to_identifier(r) }.compact
#                        end.unwrap }.compact


def name_to_fos(name: str) -> Optional[dict]:
    """Convert name to Fields of Science (OECD) subject"""
    #   # first find subject in Fields of Science (OECD)
    #   fos = JSON.load(File.read(File.expand_path('../../resources/oecd/fos-mappings.json',
    #                                              __dir__))).fetch('fosFields')

    #   subject = fos.find { |l| l['fosLabel'] == name || 'FOS: ' + l['fosLabel'] == name }

    #   if subject
    #     return [{
    #       'subject': sanitize(name).downcase
    #     },
    #             {
    #               'subject': 'FOS: ' + subject['fosLabel'],
    #               'subjectScheme': 'Fields of Science and Technology (FOS)',
    #               'schemeUri': 'http://www.oecd.org/science/inno/38235147.pdf'
    #             }]
    #   end

    #   # if not found, look in Fields of Research (Australian and New Zealand Standard Research Classification)
    #   # and map to Fields of Science. Add an extra entry for the latter
    #   fores = JSON.load(File.read(File.expand_path('../../resources/oecd/for-mappings.json',
    #                                                __dir__)))
    #   for_fields = fores.fetch('forFields')
    #   for_disciplines = fores.fetch('forDisciplines')

    #   subject = for_fields.find { |l| l['forLabel'] == name } ||
    #             for_disciplines.find { |l| l['forLabel'] == name }

    #   if subject
    #     [{
    #       'subject': sanitize(name).downcase
    #     },
    #      {
    #        'subject': 'FOS: ' + subject['fosLabel'],
    #        'subjectScheme': 'Fields of Science and Technology (FOS)',
    #        'schemeUri': 'http://www.oecd.org/science/inno/38235147.pdf'
    #      }]
    #   else

    return {"subject": name.strip()}


def encode_doi(prefix):
    """Generate a DOI using the DOI prefix and a random base32 suffix"""
    suffix = base32.generate(length=10, split_every=5, checksum=True)
    return f"https://doi.org/{prefix}/{suffix}"


def decode_doi(doi: str) -> int:
    """Decode a DOI to a number"""
    suffix = doi.split("/", maxsplit=5)[-1]
    return base32.decode(suffix)


def from_curie(id: Optional[str]) -> Optional[str]:
    """from CURIE"""
    if id is None:
        return None
    _type = id.split(":")[0]
    if _type == "DOI":
        return doi_as_url(id.split(":")[1])
    elif _type == "ROR":
        return "https://ror.org/" + id.split(":")[1]
    elif _type == "ISNI":
        return "https://isni.org/isni/" + id.split(":")[1]
    elif _type == "ORCID":
        return normalize_orcid(id.split(":")[1])
    elif _type == "URL":
        return normalize_url(id.split(":")[1])
    elif _type == "JDP":
        return id.split(":")[1]
    # TODO: resolvable url for other identifier types
    return None


def get_language(lang: str) -> Optional[dict]:
    """Provide a language object based on ISO 639, with either a name in English,
    ISO 639-1, or ISO 639-3 code as input.
    """
    if not lang:
        return None
    if len(lang) == 2:
        language = pycountry.languages.get(alpha_2=lang)
    elif len(lang) == 3:
        language = pycountry.languages.get(alpha_3=lang)
    else:
        language = pycountry.languages.get(name=lang)
    return language


def start_case(content: str) -> str:
    """Capitalize first letter of each word without lowercasing the rest"""
    words = content.split(" ")
    content = " ".join([word[0].upper() + word[1:] for word in words])
    return content


def timer_func(func):
    def function_timer(*args, **kwargs):
        start = time.time()
        value = func(*args, **kwargs)
        end = time.time()
        runtime = end - start
        msg = "{func} took {time} seconds to complete its execution."
        print(msg.format(func=func.__name__, time=runtime))
        return value

    return function_timer



"""Crossref utils module for commonmeta-py"""
from lxml import etree
from typing import Optional
from datetime import datetime
from dateutil.parser import parse
import uuid
import pydash as py_

from .constants import Commonmeta
from .utils import wrap, compact, normalize_orcid, normalize_id
from .doi_utils import doi_from_url, validate_doi


def generate_crossref_xml(metadata: Commonmeta) -> Optional[str]:
    """Generate Crossref XML. First checks for write errors (JSON schema validation)"""
    xml = crossref_root()
    head = etree.SubElement(xml, "head")
    # we use a uuid as batch_id
    etree.SubElement(head, "doi_batch_id").text = str(uuid.uuid4())
    etree.SubElement(head, "timestamp").text = datetime.now().strftime("%Y%m%d%H%M%S")
    depositor = etree.SubElement(head, "depositor")
    etree.SubElement(depositor, "depositor_name").text = metadata.depositor
    etree.SubElement(depositor, "email_address").text = metadata.email
    etree.SubElement(head, "registrant").text = metadata.registrant

    body = etree.SubElement(xml, "body")
    body = insert_crossref_work(metadata, body)
    return etree.tostring(
        xml,
        doctype='<?xml version="1.0" encoding="UTF-8"?>',
        pretty_print=True,
    )


def insert_crossref_work(metadata, xml):
    """Insert crossref work"""
    if metadata.type not in ["JournalArticle", "Article"]:
        return xml
    if doi_from_url(metadata.id) is None or metadata.url is None:
        return xml
    if metadata.type == "JournalArticle":
        xml = insert_journal(metadata, xml)
    elif metadata.type == "Article":
        xml = insert_posted_content(metadata, xml)


def insert_journal(metadata, xml):
    """Insert journal"""
    journal = etree.SubElement(xml, "journal")
    if metadata.language is not None:
        journal_metadata = etree.SubElement(
            journal, "journal_metadata", {"language": metadata.language[:2]}
        )
    else:
        journal_metadata = etree.SubElement(journal, "journal_metadata")
    if (
        metadata.container is not None
        and metadata.container.get("title", None) is not None
    ):
        etree.SubElement(journal_metadata, "full_title").text = metadata.container.get(
            "title"
        )
    journal_metadata = insert_group_title(metadata, journal_metadata)
    journal_article = etree.SubElement(
        journal, "journal_article", {"publication_type": "full_text"}
    )
    journal_article = insert_crossref_titles(metadata, journal_article)
    journal_article = insert_crossref_contributors(metadata, journal_article)
    journal_article = insert_crossref_publication_date(metadata, journal_article)
    journal_article = insert_crossref_abstract(metadata, journal_article)
    journal_article = insert_crossref_issn(metadata, journal_article)
    journal_article = insert_item_number(metadata, journal_article)
    journal_article = insert_funding_references(metadata, journal_article)
    journal_article = insert_crossref_access_indicators(metadata, journal_article)
    journal_article = insert_crossref_relations(metadata, journal_article)
    journal_article = insert_archive_locations(metadata, journal_article)
    journal_article = insert_doi_data(metadata, journal_article)
    journal_article = insert_citation_list(metadata, journal_article)

    return journal


def insert_posted_content(metadata, xml):
    """Insert posted content"""
    if metadata.language is not None:
        posted_content = etree.SubElement(
            xml, "posted_content", {"type": "other", "language": metadata.language[:2]}
        )
    else:
        posted_content = etree.SubElement(xml, "posted_content", {"type": "other"})

    posted_content = insert_group_title(metadata, posted_content)
    posted_content = insert_crossref_contributors(metadata, posted_content)
    posted_content = insert_crossref_titles(metadata, posted_content)
    posted_content = insert_posted_date(metadata, posted_content)
    posted_content = insert_institution(metadata, posted_content)
    posted_content = insert_item_number(metadata, posted_content)
    posted_content = insert_crossref_abstract(metadata, posted_content)
    posted_content = insert_funding_references(metadata, posted_content)
    posted_content = insert_crossref_access_indicators(metadata, posted_content)
    posted_content = insert_crossref_relations(metadata, posted_content)
    posted_content = insert_archive_locations(metadata, posted_content)
    posted_content = insert_doi_data(metadata, posted_content)
    posted_content = insert_citation_list(metadata, posted_content)

    return xml


def insert_group_title(metadata, xml):
    """Insert group title"""
    if metadata.subjects is None or len(metadata.subjects) == 0:
        return xml
    etree.SubElement(xml, "group_title").text = metadata.subjects[0].get(
        "subject", None
    )
    return xml


def insert_crossref_contributors(metadata, xml):
    """Insert crossref contributors"""
    if metadata.contributors is None or len(metadata.contributors) == 0:
        return xml
    contributors = etree.SubElement(xml, "contributors")
    con = [
        c
        for c in metadata.contributors
        if c.get("contributorRoles", None) == ["Author"]
        or c.get("contributorRoles", None) == ["Editor"]
    ]
    for num, contributor in enumerate(con):
        contributor_role = (
            "author" if contributor.get("contributorRoles") == ["Author"] else "editor"
        )
        sequence = "first" if num == 0 else "additional"
        if (
            contributor.get("type", None) == "Organization"
            and contributor.get("name", None) is not None
        ):
            etree.SubElement(
                contributors,
                "organization",
                {"contributor_role": contributor_role, "sequence": sequence},
            ).text = contributor.get("name")
        elif (
            contributor.get("givenName", None) is not None
            or contributor.get("familyName", None) is not None
        ):
            person_name = etree.SubElement(
                contributors,
                "person_name",
                {"contributor_role": contributor_role, "sequence": sequence},
            )
            person_name = insert_crossref_person(contributor, person_name)
        elif contributor.get("affiliation", None) is not None:
            anonymous = etree.SubElement(
                contributors,
                "anonymous",
                {"contributor_role": contributor_role, "sequence": sequence},
            )
            anonymous = insert_crossref_anonymous(contributor, anonymous)
        else:
            etree.SubElement(
                contributors,
                "anonymous",
                {"contributor_role": contributor_role, "sequence": sequence},
            )
    return xml


def insert_crossref_person(contributor, xml):
    """Insert crossref person"""
    if contributor.get("givenName", None) is not None:
        etree.SubElement(xml, "given_name").text = contributor.get("givenName")
    if contributor.get("familyName", None) is not None:
        etree.SubElement(xml, "surname").text = contributor.get("familyName")

    orcid = normalize_orcid(contributor.get("id", None))
    if orcid is not None:
        etree.SubElement(xml, "ORCID").text = orcid

    if contributor.get("affiliation", None) is None:
        return xml

    affiliations = etree.SubElement(xml, "affiliations")
    institution = etree.SubElement(affiliations, "institution")
    if py_.get(contributor, "affiliation.0.name") is not None:
        etree.SubElement(institution, "institution_name").text = py_.get(
            contributor, "affiliation.0.name"
        )
    if py_.get(contributor, "affiliation.0.id") is not None:
        etree.SubElement(institution, "institution_id", {"type": "ror"}).text = py_.get(
            contributor, "affiliation.0.id"
        )
    return xml


def insert_crossref_anonymous(contributor, xml):
    """Insert crossref anonymous"""
    if contributor.get("affiliation", None) is None:
        return xml
    affiliations = etree.SubElement(xml, "affiliations")
    institution = etree.SubElement(affiliations, "institution")
    if py_.get(contributor, "affiliation.0.name") is not None:
        etree.SubElement(institution, "institution_name").text = py_.get(
            contributor, "affiliation.0.name"
        )
    return xml


def insert_crossref_titles(metadata, xml):
    """Insert crossref titles"""
    titles = etree.SubElement(xml, "titles")
    for title in wrap(metadata.titles):
        if isinstance(title, dict):
            etree.SubElement(titles, "title").text = title.get("title", None)
        else:
            etree.SubElement(titles, "title").text = title
    return xml


def insert_citation_list(metadata, xml):
    """Insert citation list"""
    if metadata.references is None or len(metadata.references) == 0:
        return xml

    citation_list = etree.SubElement(xml, "citation_list")
    for ref in metadata.references:
        citation = etree.SubElement(
            citation_list, "citation", {"key": ref.get("key", None)}
        )
        if ref.get("journal_title", None) is not None:
            etree.SubElement(citation, "journal_article").text = ref.get(
                "journal_title"
            )
        if ref.get("author", None) is not None:
            etree.SubElement(citation, "author").text = ref.get("author")
        if ref.get("volume", None) is not None:
            etree.SubElement(citation, "volume").text = ref.get("volume")
        if ref.get("first_page", None) is not None:
            etree.SubElement(citation, "first_page").text = ref.get("first_page")
        if ref.get("publicationYear", None) is not None:
            etree.SubElement(citation, "cYear").text = ref.get("publicationYear")
        if ref.get("title", None) is not None:
            etree.SubElement(citation, "article_title").text = ref.get("title")
        if ref.get("doi", None) is not None:
            etree.SubElement(citation, "doi").text = doi_from_url(ref.get("doi"))
        if ref.get("url", None) is not None:
            etree.SubElement(citation, "unstructured_citation").text = ref.get("url")
    return xml


def insert_crossref_access_indicators(metadata, xml):
    """Insert crossref access indicators"""
    rights_uri = (
        metadata.license.get("url", None) if metadata.license is not None else None
    )
    if rights_uri is None:
        return xml
    program = etree.SubElement(
        xml,
        "program",
        {
            "xmlns": "http://www.crossref.org/AccessIndicators.xsd",
            "name": "AccessIndicators",
        },
    )
    etree.SubElement(program, "license_ref", {"applies_to": "vor"}).text = rights_uri
    etree.SubElement(program, "license_ref", {"applies_to": "tdm"}).text = rights_uri
    return xml


def insert_crossref_relations(metadata, xml):
    """Insert crossref relations"""
    if metadata.related_identifiers is None or len(metadata.related_identifiers) == 0:
        return xml
    program = etree.SubElement(
        xml,
        "program",
        {
            "xmlns": "http://www.crossref.org/relations.xsd",
            "name": "relations",
        },
    )
    for related_identifier in metadata.related_identifiers:
        related_item = etree.SubElement(program, "related_item")
        identifier_type = (
            "doi" if validate_doi(related_identifier.get("id", None)) else "uri"
        )
        _id = (
            doi_from_url(related_identifier.get("id", None))
            if identifier_type == "doi"
            else related_identifier.get("id", None)
        )
        etree.SubElement(
            related_item,
            "intra_work_relation",
            {
                "relationship-type": related_identifier.get("type", None),
                "identifier-type": identifier_type,
            },
        ).text = _id

    return xml


def insert_funding_references(metadata, xml):
    """Insert funding references"""
    if metadata.funding_references is None or len(metadata.funding_references) == 0:
        return xml
    program = etree.SubElement(
        xml,
        "program",
        {
            "xmlns": "http://www.crossref.org/fundref.xsd",
            "name": "fundref",
        },
    )
    for funding_reference in metadata.funding_references:
        assertion = etree.SubElement(program, "assertion", {"name": "fundgroup"})
        funder_name = etree.SubElement(
            assertion,
            "assertion",
            {"name": "funder_name"},
        )
        if funding_reference.get("funderIdentifier", None) is not None:
            etree.SubElement(
                funder_name,
                "assertion",
                {"name": "funder_identifier"},
            ).text = funding_reference.get("funderIdentifier", None)
        if funding_reference.get("awardNumber", None) is not None:
            etree.SubElement(
                assertion,
                "assertion",
                {"name": "award_number"},
            ).text = funding_reference.get("awardNumber", None)
        funder_name.text = funding_reference.get("funderName", None)
    return xml


def insert_crossref_subjects(metadata, xml):
    """Insert crossref subjects"""
    if metadata.subjects is None:
        return xml
    subjects = etree.SubElement(xml, "subjects")
    for subject in metadata.subjects:
        if isinstance(subject, dict):
            etree.SubElement(subjects, "subject").text = subject.get("subject", None)
        else:
            etree.SubElement(subjects, "subject").text = subject
    return xml


def insert_crossref_language(metadata, xml):
    """Insert crossref language"""
    if metadata.language is None:
        return xml
    etree.SubElement(xml, "language").text = metadata.language
    return xml


def insert_crossref_publication_date(metadata, xml):
    """Insert crossref publication date"""
    pub_date = parse(metadata.date.get("published", None))
    if pub_date is None:
        return xml

    publication_date = etree.SubElement(
        xml, "publication_date", {"media_type": "online"}
    )
    etree.SubElement(publication_date, "month").text = f"{pub_date.month:d}"
    etree.SubElement(publication_date, "day").text = f"{pub_date.day:d}"
    etree.SubElement(publication_date, "year").text = str(pub_date.year)
    return xml


def insert_posted_date(metadata, xml):
    """Insert posted date"""
    pub_date = parse(metadata.date.get("published", None))
    if pub_date is None:
        return xml

    posted_date = etree.SubElement(xml, "posted_date", {"media_type": "online"})
    etree.SubElement(posted_date, "month").text = f"{pub_date.month:d}"
    etree.SubElement(posted_date, "day").text = f"{pub_date.day:d}"
    etree.SubElement(posted_date, "year").text = str(pub_date.year)
    return xml


def insert_institution(metadata, xml):
    """Insert institution"""
    if metadata.publisher.get("name", None) is None:
        return xml
    institution = etree.SubElement(xml, "institution")
    etree.SubElement(institution, "institution_name").text = metadata.publisher.get(
        "name"
    )
    return xml


def insert_item_number(metadata, xml):
    """Insert item number"""
    if metadata.alternate_identifiers is None:
        return xml
    for alternate_identifier in metadata.alternate_identifiers:
        if alternate_identifier.get("alternateIdentifier", None) is None:
            continue
        if alternate_identifier.get("alternateIdentifierType", None) is not None:
            # strip hyphen from UUIDs, as item_number can only be 32 characters long (UUIDv4 is 36 characters long)
            if alternate_identifier.get("alternateIdentifierType", None) == "UUID":
                alternate_identifier["alternateIdentifier"] = alternate_identifier.get(
                    "alternateIdentifier", ""
                ).replace("-", "")
            etree.SubElement(
                xml,
                "item_number",
                {
                    "item_number_type": alternate_identifier.get(
                        "alternateIdentifierType", ""
                    ).lower()
                },
            ).text = alternate_identifier.get("alternateIdentifier", None)
        else:
            etree.SubElement(xml, "item_number").text = alternate_identifier.get(
                "alternateIdentifier", None
            )
    return xml


def insert_archive_locations(metadata, xml):
    """Insert archive locations"""
    if metadata.archive_locations is None:
        return xml
    archive_locations = etree.SubElement(xml, "archive_locations")
    for archive_location in metadata.archive_locations:
        etree.SubElement(archive_locations, "archive", {"name": archive_location})
    return xml


def insert_doi_data(metadata, xml):
    """Insert doi data"""
    if doi_from_url(metadata.id) is None or metadata.url is None:
        return xml
    doi_data = etree.SubElement(xml, "doi_data")
    etree.SubElement(doi_data, "doi").text = doi_from_url(metadata.id)
    etree.SubElement(doi_data, "resource").text = metadata.url
    collection = etree.SubElement(doi_data, "collection", {"property": "text-mining"})
    item = etree.SubElement(collection, "item")
    etree.SubElement(item, "resource", {"mime_type": "text/html"}).text = metadata.url
    if metadata.files is None:
        return xml
    for file in metadata.files:
        # Crossref schema currently doesn't support text/markdown
        if file.get("mimeType", None) == "text/markdown":
            file["mimeType"] = "text/plain"
        item = etree.SubElement(collection, "item")
        etree.SubElement(
            item, "resource", {"mime_type": file.get("mimeType", "")}
        ).text = file.get("url", None)
    return xml


def insert_crossref_license(metadata, xml):
    """Insert crossref license"""
    if metadata.license is None:
        return xml
    license_ = etree.SubElement(xml, "license")
    if isinstance(metadata.license, dict):
        r = metadata.license
    else:
        r = {}
        r["rights"] = metadata.license
        r["rightsUri"] = normalize_id(metadata.license)
    attributes = compact(
        {
            "rightsURI": r.get("rightsUri", None),
            "rightsIdentifier": r.get("rightsIdentifier", None),
            "rightsIdentifierScheme": r.get("rightsIdentifierScheme"),
            "schemeURI": r.get("schemeUri", None),
            "xml:lang": r.get("lang", None),
        }
    )
    etree.SubElement(license_, "rights", attributes).text = r.get("rights", None)
    return xml


def insert_crossref_issn(metadata, xml):
    """Insert crossref issn"""
    if (
        metadata.container is None
        or metadata.container.get("identifierType", None) != "ISSN"
    ):
        return xml
    etree.SubElement(xml, "issn").text = metadata.container.get("identifier", None)
    return xml


def insert_crossref_abstract(metadata, xml):
    """Insert crossref abstrac"""
    if metadata.descriptions is None:
        return xml
    if isinstance(metadata.descriptions[0], dict):
        d = metadata.descriptions[0]
    else:
        d = {}
        d["description"] = metadata.descriptions[0]
    abstract = etree.SubElement(
        xml, "abstract", {"xmlns": "http://www.ncbi.nlm.nih.gov/JATS1"}
    )
    etree.SubElement(abstract, "p").text = d.get("description", None)
    return xml


def crossref_root():
    """Crossref root with namespaces"""
    doi_batch = """<doi_batch xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns="http://www.crossref.org/schema/5.3.1" xmlns:jats="http://www.ncbi.nlm.nih.gov/JATS1" xmlns:fr="http://www.crossref.org/fundref.xsd" xmlns:mml="http://www.w3.org/1998/Math/MathML" xsi:schemaLocation="http://www.crossref.org/schema/5.3.1 https://www.crossref.org/schemas/crossref5.3.1.xsd" version="5.3.1"></doi_batch>"""
    return etree.fromstring(doi_batch)


def generate_crossref_xml_list(metalist) -> Optional[str]:
    """Generate Crossref XML list."""
    if not metalist.is_valid:
        return None
    xml = crossref_root()
    head = etree.SubElement(xml, "head")
    # we use a uuid as batch_id
    etree.SubElement(head, "doi_batch_id").text = str(uuid.uuid4())
    etree.SubElement(head, "timestamp").text = datetime.now().strftime("%Y%m%d%H%M%S")
    depositor = etree.SubElement(head, "depositor")
    etree.SubElement(depositor, "depositor_name").text = metalist.depositor or "test"
    etree.SubElement(depositor, "email_address").text = (
        metalist.email or "info@example.org"
    )
    etree.SubElement(head, "registrant").text = metalist.registrant or "test"

    body = etree.SubElement(xml, "body")
    body = [insert_crossref_work(item, body) for item in metalist.items]
    return etree.tostring(
        xml,
        doctype='<?xml version="1.0" encoding="UTF-8"?>',
        pretty_print=True,
    )



import click
import time
import pydash as py_
import orjson as json

from commonmeta import Metadata, MetadataList  # __version__
from commonmeta.api_utils import update_ghost_post_via_api
from commonmeta.doi_utils import validate_prefix
from commonmeta.utils import encode_doi, decode_doi
from commonmeta.readers.json_feed_reader import (
    get_json_feed_item_uuid,
)
from commonmeta.readers.crossref_reader import get_random_crossref_id
from commonmeta.readers.datacite_reader import get_random_datacite_id


@click.group()
@click.option("--show-errors", default=False)
def cli(show_errors):
    if show_errors:
        click.echo("Show errors mode is on")


@cli.command()
@click.argument("input", type=str, required=True)
@click.option("--via", "-f", type=str, default=None)
@click.option("--to", "-t", type=str, default="commonmeta")
@click.option("--style", "-s", type=str, default="apa")
@click.option("--locale", "-l", type=str, default="en-US")
@click.option("--doi", type=str)
@click.option("--prefix", type=str)
@click.option("--depositor", type=str)
@click.option("--email", type=str)
@click.option("--registrant", type=str)
@click.option("--show-errors/--no-errors", type=bool, show_default=True, default=False)
def convert(
    input,
    via,
    to,
    style,
    locale,
    doi,
    prefix,
    depositor,
    email,
    registrant,
    show_errors,
):
    metadata = Metadata(input, via=via, doi=doi, prefix=prefix)
    if show_errors and not metadata.is_valid:
        raise click.ClickException(str(metadata.errors) + str(metadata.write_errors))

    click.echo(
        metadata.write(
            to=to,
            style=style,
            locale=locale,
            depositor=depositor,
            email=email,
            registrant=registrant,
        )
    )
    if show_errors and metadata.write_errors:
        raise click.ClickException(str(metadata.write_errors))


@cli.command()
@click.argument("string", type=str, required=True)
@click.option("--via", "-f", type=str)
@click.option("--to", "-t", type=str, default="commonmeta")
@click.option("--style", "-s", type=str, default="apa")
@click.option("--locale", "-l", type=str, default="en-US")
@click.option("--prefix", type=str)
@click.option("--depositor", type=str)
@click.option("--email", type=str)
@click.option("--registrant", type=str)
@click.option("--filename", type=str)
@click.option("--jsonlines/--no-jsonlines", type=bool, show_default=True, default=False)
@click.option("--show-errors/--no-errors", type=bool, show_default=True, default=False)
@click.option("--show-timer/--no-timer", type=bool, show_default=True, default=False)
def list(
    string,
    via,
    to,
    style,
    locale,
    prefix,
    depositor,
    email,
    registrant,
    filename,
    jsonlines,
    show_errors,
    show_timer,
):
    start = time.time()
    metadata_list = MetadataList(
        string,
        via=via,
        depositor=depositor,
        email=email,
        registrant=registrant,
        prefix=prefix,
        filename=filename,
        jsonlines=jsonlines,
    )
    end = time.time()
    runtime = end - start
    if show_errors and not metadata_list.is_valid:
        raise click.ClickException(str(metadata_list.errors))
    click.echo(metadata_list.write(to=to, style=style, locale=locale))
    if show_errors and len(metadata_list.write_errors) > 0:
        raise click.ClickException(str(metadata_list.write_errors))
    if show_timer:
        click.echo(f"Runtime: {runtime:.2f} seconds")


@cli.command()
@click.option("--provider", type=str, default="crossref")
@click.option("--prefix", type=str)
@click.option("--type", type=str)
@click.option("--number", "-n", type=int, default=1)
@click.option("--to", "-t", type=str, default="commonmeta")
@click.option("--style", "-s", type=str, default="apa")
@click.option("--locale", "-l", type=str, default="en-US")
@click.option("--show-errors/--no-errors", type=bool, show_default=True, default=False)
def sample(provider, prefix, type, number, to, style, locale, show_errors):
    if provider == "crossref":
        string = json.dumps(
            {"items": get_random_crossref_id(number, prefix=prefix, _type=type)}
        )
    elif provider == "datacite":
        string = json.dumps({"items": get_random_datacite_id(number)})
    else:
        output = "Provider not supported. Use 'crossref' or 'datacite' instead."
        click.echo(output)
    lst = MetadataList(
        string,
        via=provider,
        style=style,
        locale=locale,
    )
    for item in lst.items:
        output = item.write(to=to)
        if show_errors and not item.is_valid:
            message = f"{item}: {item.errors}"
            raise click.ClickException(message)
        click.echo(output)


@cli.command()
@click.argument("prefix", type=str, required=True)
def encode(prefix):
    if validate_prefix(prefix) is None:
        return None
    output = encode_doi(prefix)
    click.echo(output)


@cli.command()
@click.argument("doi", type=str, required=True)
def decode(doi):
    output = decode_doi(doi)
    click.echo(output)


@cli.command()
@click.argument("id", type=str, required=True)
def encode_by_id(id):
    post = get_json_feed_item_uuid(id)
    prefix = py_.get(post, "blog.prefix")
    if validate_prefix(prefix) is None:
        return None
    output = encode_doi(prefix)
    click.echo(output)


@cli.command()
@click.argument("filter", type=str, required=True, default="unregistered")
@click.option("--id", type=str)
def json_feed(filter, id=None):
    if filter == "blog_slug" and id is not None:
        post = get_json_feed_item_uuid(id)
        output = py_.get(post, "blog.slug", "no slug found")
    else:
        output = "no filter specified"
    click.echo(output)


@cli.command()
@click.argument("id", type=str, required=True)
@click.option("--api-key", "-k", type=str, required=True)
@click.option("--api-url", "-u", type=str, required=True)
def update_ghost_post(id, api_key, api_url):
    output = update_ghost_post_via_api(id, api_key, api_url)
    click.echo(output)


if __name__ == "__main__":
    cli()



"""Schema.org writer for commonmeta-py"""
import orjson as json
from ..utils import to_schema_org_creators, github_as_repo_url
from ..base_utils import compact, wrap, presence, parse_attributes
from ..constants import CM_TO_SO_TRANSLATIONS


def write_schema_org(metadata):
    """Write schema.org"""
    container = metadata.container
    if metadata.type == "Dataset" and metadata.files is not None:
        media_objects = [
            compact(
                {
                    "@type": "DataDownload",
                    "contentUrl": file.get("url"),
                    "encodingFormat": file.get("mimeType", None),
                    "name": file.get("key", None),
                    "sha256": file["checksum"]
                    if file.get("checksum", None)
                    and file["checksum"].startswith("sha256")
                    else None,
                    "size": file.get("size", None),
                }
            )
            for file in metadata.files
        ]
    elif metadata.files is not None:
        media_objects = [
            compact(
                {
                    "@type": "MediaObject",
                    "contentUrl": file.get("url"),
                    "encodingFormat": file.get("mimeType", None),
                    "name": file.get("key", None),
                    "sha256": file["checksum"]
                    if file.get("checksum", None)
                    and file["checksum"].startswith("sha256")
                    else None,
                    "size": file.get("size", None),
                }
            )
            for file in metadata.files
        ]
    else:
        media_objects = None
    if metadata.type == "Dataset" and container is not None:
        data_catalog = compact(
            {
                "@id": container.get("id", None),
                "@type": "DataCatalog",
                "name": container.get("title", None),
            }
        )
        periodical = None
    elif container is not None:
        periodical = compact(
            {
                "issn": container.get("identifier", None)
                if container.get("identifierType", None) == "ISSN"
                else None,
                "@id": container.get("identifier", None)
                if container.get("identifierType", None) != "ISSN"
                else None,
                "@type": container.get("type", None)
                if container.get("type", None) == "Journal"
                else None,
                "additionalType": container.get("type", None)
                if container.get("type", None) != "Journal"
                else None,
                "name": container.get("title", None),
            }
        )
        data_catalog = None
    else:
        periodical = None
        data_catalog = None
    schema_org = CM_TO_SO_TRANSLATIONS.get(metadata.type, "CreativeWork")
    additional_type = metadata.additional_type
    authors = [
        au for au in wrap(metadata.contributors) if au["contributorRoles"] == ["Author"]
    ]
    editors = [
        au for au in wrap(metadata.contributors) if au["contributorRoles"] == ["Editor"]
    ]
    if metadata.type == "Software":
        rel = next(
            (
                relation
                for relation in metadata.related_identifiers
                if relation["type"] == "IsSupplementTo"
            ),
            None,
        )
        code_repository = github_as_repo_url(rel["id"])
    else:
        code_repository = None

    data = compact(
        {
            "@context": "http://schema.org",
            "@id": metadata.id,
            "identifier": metadata.id,
            "@type": schema_org,
            "url": metadata.url,
            "additionalType": additional_type,
            "name": parse_attributes(metadata.titles, content="title", first=True),
            "author": to_schema_org_creators(authors),
            "editor": to_schema_org_creators(editors),
            "description": parse_attributes(
                metadata.descriptions, content="description", first=True
            ),
            "license": metadata.license.get("url", None) if metadata.license else None,
            "version": metadata.version,
            "keywords": presence(
                parse_attributes(
                    wrap(metadata.subjects), content="subject", first=False
                )
            ),
            "inLanguage": metadata.language,
            "dateCreated": metadata.date.get("created", None),
            "datePublished": metadata.date.get("published", None),
            "dateModified": metadata.date.get("updated", None),
            "pageStart": container.get("firstPage", None) if container else None,
            "pageEnd": container.get("lastPage", None) if container else None,
            # "isPartOf": unwrap(to_schema_org_relations(
            #     related_items=metadata.related_items,
            #     relation_type="IsPartOf",
            # )),
            "periodical": periodical if periodical else None,
            "includedInDataCatalog": data_catalog if data_catalog else None,
            "distribution": media_objects if metadata.type == "Dataset" else None,
            "encoding": media_objects if metadata.type != "Dataset" else None,
            "codeRepository": code_repository,
            "publisher": {
                "@type": "Organization",
                "name": metadata.publisher.get("name", None),
            }
            if metadata.publisher
            else None,
            "provider": {"@type": "Organization", "name": metadata.provider}
            if metadata.provider
            else None,
        }
    )
    return json.dumps(data)



"""Commonmeta writer for commonmeta-py"""
import orjson as json
import orjsonl
from ..base_utils import compact


def write_commonmeta(metadata):
    """Write commonmeta"""
    if metadata is None:
        return None

    data = compact(
        {
            # required properties
            "id": metadata.id,
            "type": metadata.type,
            "url": metadata.url,
            "contributors": metadata.contributors,
            "titles": metadata.titles,
            "publisher": metadata.publisher,
            "date": metadata.date,
            # recommended and optional properties
            "container": metadata.container,
            "subjects": metadata.subjects,
            "language": metadata.language,
            "references": metadata.references,
            "sizes": metadata.sizes,
            "formats": metadata.formats,
            "version": metadata.version,
            "license": metadata.license,
            "descriptions": metadata.descriptions,
            "geo_locations": metadata.geo_locations,
            "alternate_identifiers": metadata.alternate_identifiers,
            "related_identifiers": metadata.related_identifiers,
            "funding_references": metadata.funding_references,
            # other properties
            "files": metadata.files,
            "provider": metadata.provider,
        }
    )
    return json.dumps(data)


def write_commonmeta_list(metalist):
    """Write commonmeta list. If filename is provided,
    write to file. Optionally, use JSON Lines format."""
    if metalist is None:
        return None

    items = [compact(vars(item)) for item in metalist.items]

    if metalist.filename and metalist.filename.rsplit('.', 1)[1] in ["jsonl", "json"]:
        if metalist.jsonlines:
            orjsonl.save(metalist.filename, items)
        else:
            json_items = json.dumps(items).decode("utf-8")
            with open(metalist.filename, "w") as file:
                file.write(json_items)
        return metalist.filename
    else:
        return json.dumps(items)
